/*
 * Copyright (C) 2009
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: Autonomous Technologies
 * file .......: rtcBinaryInputHandler.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 01/20/2009
 * modified ...: $Date: 2009-01-21 18:19:16 -0800 (Wed, 21 Jan 2009) $
 * changed by .: $Author: benjaminpitzer $
 * revision ...: $Revision: 14 $
 */
#ifndef RTC_BINARY_INPUT_HANDLER_H
#define RTC_BINARY_INPUT_HANDLER_H

//== INCLUDES ==================================================================
#include "rtc/rtcInputHandler.h"

//== NAMESPACES ================================================================
namespace rtc {

/**
 * Binary input handler.
 *
 * The BinaryInputHandler class offers an interface for the functor
 * parameters and other classes to read() them in a binary format.
 *
 * The advantage of the binary format is that the produced files are
 * smaller, which at the same time implies shorter times to read or write
 * the data.
 *
 * The disadvange is that the produced files cannot be easily edited.  If
 * you need this, you should still use the rtc::XmlOutputHandler.
 *
 * \warning Do not forget to open your streams in binary mode.  If you
 *          create them as ASCII streams, the data can be corrupted.
 *
 * Example:
 * \code
 * // the binary stream formatting object
 * rtc::BinaryInputHandler bsh;
 *
 * // Read example
 *
 * // Open a stream in binary mode
 * std::ifstream in("testfile.bin",std::ios_base::binary);
 *
 * bsh.use(in);
 * int i;
 * rtc::read(bsh,i);
 *
 * std::string str;
 * rtc::read(bsh,str);
 *
 * in.close();
 * \endcode
 */
class BinaryInputHandler : public InputHandler {
public:
  /// default constructor
  BinaryInputHandler();
  /// construct from istream
  BinaryInputHandler(std::istream& is);
  /// construct from filename
  BinaryInputHandler(const char* filename);
  /// destructor
  virtual ~BinaryInputHandler();
  /// returns if the data format is binary or ascii
  virtual bool binary() const;
  /// read the name of a name-data pair
  virtual bool readName(const std::string& name);
  /// read the separator of a name-data pair
  virtual bool readSeparator();
  //@{
  /**
   * read members for symbol-data pairs
   */
  virtual bool read(std::string& data);
  virtual bool read(double& data);
  virtual bool read(float& data);
  virtual bool read(int& data);
  virtual bool read(unsigned int& data);
  virtual bool read(char& data);
  virtual bool read(unsigned char& data);
  virtual bool read(bool& data);
  virtual bool read(long& data);
  virtual bool read(unsigned long& data);
  virtual bool read(short& data);
  virtual bool read(unsigned short& data);
  virtual bool read(char* data, long n);
  //@}
};

//==============================================================================
} // namespace rtc
//==============================================================================
#endif // RTC_BINARY_INPUT_HANDLER_H defined
//==============================================================================

