/*
 * Copyright (C) 2009
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: Autonomous Technologies
 * file .......: rtcOutputHandler.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 01/20/2009
 * modified ...: $Date: 2009-01-21 18:19:16 -0800 (Wed, 21 Jan 2009) $
 * changed by .: $Author: benjaminpitzer $
 * revision ...: $Revision: 14 $
 */
#ifndef RTC_BINARY_OUTPUT_HANDLER_H
#define RTC_BINARY_OUTPUT_HANDLER_H

//== INCLUDES ==================================================================
#include "rtc/rtcOutputHandler.h"

//== NAMESPACES ================================================================
namespace rtc {

/**
 * Binary output handler.
 *
 * The BinaryOutputHandler class offers an interface for the functor
 * parameters and other classes to write() them in a
 * binary format.
 *
 * The advantage of the binary format is that the produced files are
 * smaller, which at the same time implies shorter times to write or write
 * the data.
 *
 * The disadvantage is that the produced files cannot be easily edited.  If
 * you need this, you should still use the rtc::XmlOutputHandler.
 *
 * \warning Do not forget to open your streams in binary mode.  If you
 *          create them as ASCII streams, the data can be corrupted.
 *
 * Example:
 * \code
 * // the binary stream formatting object
 * rtc::BinaryOutputHandler bsh;
 *
 * // Write example:
 *
 * // open a stream in binary mode
 * std::ofstream out("testfile.bin",std::ios_base::binary);
 *
 * // tell the binary stream handler to use the given stream
 * bsh.use(out);
 *
 * rtc::write(bsh,5);
 * rtc::write(bsh,"hello world");
 *
 * out.close();
 * \endcode
 */
class BinaryOutputHandler : public OutputHandler {
public:
  /// default constructor
  BinaryOutputHandler();
  /// construct from ostream
  BinaryOutputHandler(std::ostream& os);
  /// destructor
  virtual ~BinaryOutputHandler();
  /// returns if the data format is binary or ascii
  virtual bool binary() const;
  /// write the name of a name-data pair
  virtual bool writeName(const std::string& name);
  /// write the separator of a name-data pair
  virtual bool writeSeparator();
  /// write the EOL symbol
  virtual bool writeEOL();
  //@{
  /**
   * write members for symbol-data pairs
   */
  virtual bool write(const std::string& data);
  virtual bool write(const double& data);
  virtual bool write(const float& data);
  virtual bool write(const int& data);
  virtual bool write(const unsigned int& data);
  virtual bool write(const char& data);
  virtual bool write(const unsigned char& data);
  virtual bool write(const bool& data);
  virtual bool write(const long& data);
  virtual bool write(const unsigned long& data);
  virtual bool write(const short& data);
  virtual bool write(const unsigned short& data);
  virtual bool write(const char* data, long n);
  //@}
};

//==============================================================================
} // namespace rtc
//==============================================================================
#endif // RTC_BINARY_OUTPUT_HANDLER_H defined
//==============================================================================

