/*
 * Copyright (C) 2008
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: PUMA: Probablistic Unsupervised Model Acquisition
 * file .......: Functor.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 08/16/2006
 * modified ...: $Date: 2008-11-06 18:14:42 -0800 (Thu, 06 Nov 2008) $
 * changed by .: $Author: benjaminpitzer $
 * revision ...: $Revision: 558 $
 */
#ifndef FUNCTOR_H
#define FUNCTOR_H

//== INCLUDES ==================================================================
#include "rtc/rtcIOObject.h"
#include <string>
#include <vector>

//== NAMESPACES ================================================================
namespace rtc {

/**
 * Base class for all functors.
 *
 * Every functor must have at least the member <em>apply()</em>,
 * which "applies" the functor's functionality on the data given through the
 * arguments of the apply method.
 *
 * The operation of the functor can be controled with some parameters,
 * which will can be set with the "setParameters" member-function.
 */
class Functor : public IOObject {
public:
  /**
  * A status message is just a string
  */
  typedef std::string StatusMessage;

  /**
  * A list of strings as status messages
  */
  typedef std::vector<StatusMessage> StatusMessages;

  /**
   * Base class for all parameter objects
   */
  class Parameters : public IOObject {
  public:
    /**
     * default constructor
     */
    Parameters();

    /**
     * copy constructor
     */
    Parameters(const Parameters& other);

    /**
     * destructor
     */
    virtual ~Parameters();

    /**
     * copy data of "other" Parameters
     */
    Parameters& copy(const Parameters& other);

    /**
     * returns a pointer to a clone of the Parameters.
     */
    virtual Parameters* clone() const = 0;

    /**
    * read the parameters
    */
    virtual bool write(OutputHandler &oh) const;

    /**
    * read the parameters
    */
    virtual bool read(InputHandler &ih);

  protected:
    /**
     * copy data of "other" Parameters
     */
    Parameters& operator=(const Parameters& other);
  };

  /**
   * default constructor
   */
  Functor();

  /**
   * copy constructor
   */
  Functor(const Functor& other);

  /**
   * destructor
   */
  virtual ~Functor();

  /**
   * Set Functor's Parameters.
   *
   */
  virtual bool setParameters(const Parameters& theParam);

  /**
  * returns current Parameters.
  */
  const Parameters& getParameters() const;

  /**
   * copy data of "other" Functor.
   * Please note that the status string will _NOT_ be copied!
   */
  Functor& copy(const Functor& other);

  /**
   * clone member
   */
  virtual Functor* clone() const = 0;

  /**
  * Adds a status message to the messages
  */
  void addStatusMessage(const StatusMessage msg);

  /**
  * Returns the list of all status messages
  */
  const StatusMessages& getStatusMessages() const;

  /**
  * read the functor parameters
  */
  virtual bool write(OutputHandler &oh) const;

  /**
  * read the functor parameters
  */
  virtual bool read(InputHandler &ih);

private:

  /**
   * This is private to avoid default implementation.
   *
   */
  Functor& operator=(const Functor& other);

  /**
   * current Parameters.
   */
  Parameters* params;

  /**
  * current Parameters.
  */
  StatusMessages m_messages;
};

//==============================================================================
} // NAMESPACE rtc
//==============================================================================
#endif // FUNCTOR_H defined
//==============================================================================

