/*
 * Copyright (C) 2008
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: PUMA: Probablistic Unsupervised Model Acquisition
 * file .......: Geometry.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 03/28/2008
 * modified ...: $Date: 2008-08-14 19:00:23 -0700 (Thu, 14 Aug 2008) $
 * changed by .: $Author: wg75pal $
 * revision ...: $Revision: 344 $
 */
#ifndef GEOMETRY3D_H
#define GEOMETRY3D_H

//== INCLUDES ==================================================================
#include <rtc/rtcVec3.h>

namespace rtc {

// two vectors, a and b, starting from c
float dot(const Vec3f &a, const Vec3f &b);
// two vectors, a and b
Vec3f cross(const Vec3f &a, const Vec3f &b);
// two vectors, a and b, starting from c
Vec3f cross(const Vec3f &a, const Vec3f &b, const Vec3f &c);
float dist2(const Vec3f &a, const Vec3f &b);
float dist(const Vec3f &a, const Vec3f &b);
// linear interpolation
Vec3f lerp(float t, const Vec3f &a, const Vec3f &b);
// is the point centered at c within the box centered at bc, with radius br?
bool point_within_bounds(const Vec3f &c, const Vec3f &bc, float br);
// is the ball centered at b with radius r
// fully within the box centered at bc, with radius br?
bool ball_within_bounds(const Vec3f &b, float r,
                        const Vec3f &bc, float br);
// is the ball centered at b with radius r
// fully within the box centered from min to max?
bool ball_within_bounds(const Vec3f &b, float r,
                        const Vec3f &min,
                        const Vec3f &max);
// does the ball centered at b, with radius r,
// intersect the box centered at bc, with radius br?
bool bounds_overlap_ball(const Vec3f &b, float r,
                         const Vec3f &bc, float br);
bool bounds_overlap_ball(const Vec3f &b, float r,
                         const Vec3f &min, const Vec3f &max);
// calculate barycentric coordinates of the point p
// (already on the triangle plane) with normal vector n
// and two edge vectors v1 and v2,
// starting from a common vertex t0
void bary_fast(const Vec3f& p, const Vec3f& n,
               const Vec3f &t0, const Vec3f& v1,
               const Vec3f& v2, float &b1, float &b2, float &b3);
bool closer_on_lineseg(const Vec3f &x, Vec3f &cp, const Vec3f &a,
                       const Vec3f &b, float &d2);
void distance_point_line(const Vec3f &x, const Vec3f &a,
                         const Vec3f &b, float &d2, Vec3f &cp);
void distance_point_tri(const Vec3f &x, const Vec3f &t1,
                        const Vec3f &t2, const Vec3f &t3,
                        float &d2, Vec3f &cp);
bool closer_on_tri(const Vec3f &x, Vec3f &cp,
                   const Vec3f &t1, const Vec3f &t2,
                   const Vec3f &t3, float &d2);
// calculate the intersection of a line going through p
// to direction dir with a plane spanned by t1,t2,t3
// (modified from Graphics Gems, p.299)
bool line_plane_X(const Vec3f& p, const Vec3f& dir,
                  const Vec3f& t1, const Vec3f& t2,
                  const Vec3f& t3,
                  Vec3f &x, float &dist);
bool line_plane_X(const Vec3f& p, const Vec3f& dir,
                  const Vec3f& nrm, float d, Vec3f &x, float &dist);
// calculate barycentric coordinates of the point p
// on triangle t1 t2 t3
void bary(const Vec3f& p,
          const Vec3f& t1, const Vec3f& t2, const Vec3f& t3,
          float &b1, float &b2, float &b3);
// calculate barycentric coordinates for the intersection of
// a line starting from p, going to direction dir, and the plane
// of the triangle t1 t2 t3
bool bary(const Vec3f& p, const Vec3f& dir,
          const Vec3f& t1, const Vec3f& t2, const Vec3f& t3,
          float &b1, float &b2, float &b3);
// calculate the intersection of a line starting from p,
// going to direction dir, and the triangle t1 t2 t3
bool line_tri_X(const Vec3f& p, const Vec3f& dir,
                const Vec3f& t1, const Vec3f& t2, const Vec3f& t3,
                Vec3f& x, float& d);

bool closer_on_line(const Vec3f &x, const Vec3f &a, const Vec3f &b, float &d2, Vec3f &cp);
void dist_to_line(const Vec3f &x, const Vec3f &a, const Vec3f &b, float &d, Vec3f &cp);
float dist_to_line(const Vec3f &x, const Vec3f &a, const Vec3f &b);

// calculates the area of an triangle (t1 t2 t3)
template <class T> inline T rtc_triangle_area(const Vec3<T>& t1, const Vec3<T>& t2, const Vec3<T>& t3)
{
  const Vec3<T> v1 = t2-t1;
  const Vec3<T> v2 = t3-t1;
  const Vec3<T> t = v1.cross(v2);
  return T(0.5)*t.norm();
}

//=============================================================================
} // namespace rtc
//=============================================================================
#endif
