/*
 * Copyright (C) 2009
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: Autonomous Technologies
 * file .......: rtcInputHandler.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 01/20/2009
 * modified ...: $Date: 2009-03-20 11:36:51 -0700 (Fri, 20 Mar 2009) $
 * changed by .: $Author: wg75pal $
 * revision ...: $Revision: 105 $
 */
#ifndef RTC_INPUT_HANDLER_H
#define RTC_INPUT_HANDLER_H

//== INCLUDES ==================================================================
#include <string>
#include <iostream>

//== NAMESPACES ================================================================
namespace rtc {

/**
 * Base input handler class.
 *
 * The input handler classes offer an abstract interface for the IOObject
 * classes to read() them in different formats.
 *
 * All InputHandlers are used as wrappers for std::istreams objects.  Therefore,
 * the abstract interface provides the virtual methods use() to indicate which
 * stream has to be used.
 */
class InputHandler {
public:
  /// default constructor
  InputHandler();
  /// construct from istream
  InputHandler(std::istream& is);
  /// destructor
  virtual ~InputHandler();
  /// returns if the data format is binary or ascii
  virtual bool binary() const = 0;
  /// indicate the input stream to be used
  virtual void use(std::istream& is);
  /// read the name of a name-data pair
  virtual bool readName(const std::string& name) = 0;
  /// read the separator of a name-data pair
  virtual bool readSeparator() = 0;
  //@{
  /**
   * read members for symbol-data pairs
   */
  virtual bool read(std::string& data) = 0;
  virtual bool read(double& data) = 0;
  virtual bool read(float& data) = 0;
  virtual bool read(int& data) = 0;
  virtual bool read(unsigned int& data) = 0;
  virtual bool read(char& data) = 0;
  virtual bool read(unsigned char& data) = 0;
  virtual bool read(bool& data) = 0;
  virtual bool read(long& data) = 0;
  virtual bool read(unsigned long& data) = 0;
  virtual bool read(short& data) = 0;
  virtual bool read(unsigned short& data) = 0;
  virtual bool read(char* data, long n) = 0;
  //@}
  /// returns a reference to the stream
  std::istream& stream();
  /// Check if the state of the stream is good for i/o operations.
  bool good() const;
protected:
  /// pointer to the input stream
  std::istream* inStream;
};

/**
 * handler functions with standard storable interface
 */
template <class T>
bool rtc_read(InputHandler& ih, const std::string& name, T& data)
{
  bool res = true;
  res = res && ih.readName(name);
  res = res && ih.readSeparator();
  if(ih.binary())
    res = res && ih.read(data);
  else
    ih.stream() >> data;
  return res;
};

//==============================================================================
} // namespace rtc
//==============================================================================
#endif // RTC_INPUT_HANDLER_H defined
//==============================================================================

