/*
 * Copyright (C) 2007
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: PUMA: Probablistic Unsupervised Model Acquisition
 * file .......: Mesh3D.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 02/29/2008
 * modified ...: $Date:2008-03-03 10:26:02 -0800 (Mon, 03 Mar 2008) $
 * changed by .: $Author:benjaminpitzer $
 * revision ...: $Revision:141 $
 */
#ifndef MESH3D_H_
#define MESH3D_H_

//== INCLUDES ==================================================================
#include <rtc/rtcBase.h>
#include <rtc/rtcIOObject.h>
#include <rtc/rtcPose3D.h>
#include <rtc/rtcImage.h>
#ifndef NO_GRAPHICS
#include <rtc/rtcTexture.h>
#endif
#include "rtc/rtcVertex3D.h"
#include "rtc/rtcFace3D.h"

#include "sensor_msgs/CameraInfo.h"

//== NAMESPACES ================================================================
namespace rtc {

#define MESH_FLAG_HAS_VERTEX_NORMAL  0x01
#define MESH_FLAG_HAS_FACE_NORMAL    0x02
#define MESH_FLAG_HAS_VERTEX_COLOR   0x04
#define MESH_FLAG_HAS_FACE_COLOR     0x08
#define MESH_FLAG_HAS_VERTEX_FLAG    0x10
#define MESH_FLAG_HAS_FACE_FLAG      0x20
#define MESH_FLAG_HAS_TEXTURE        0x40
#define MESH_FLAG_HAS_TEXTUREMASK    0x80
#define MESH_FLAG_HAS_IMAGE_POSE     0x81
#define MESH_FLAG_HAS_CAMERA_INFO    0x82
#define MESH_FLAG_HAS_FRAME_ID       0x83

typedef std::vector<Vertex3D*> VecVertex3D;
typedef std::vector<Face3D*> VecFace3D;

//== CLASS DEFINITION ==========================================================
class Mesh3D : public IOObject
{
public:
  // standard c'tor and d'tor
  Mesh3D();
  virtual ~Mesh3D();
  // copy c'tor
  Mesh3D(const Mesh3D& other);
  const Mesh3D &operator=(const Mesh3D& other);

  // performs updateValid, updateBoundary, and updateNormals
  void update();
  // calculates normals for valid and non-boundary points
  void updateNormals();
  // clear all edges and vertices
  void clear();
  void set(const Mesh3D& other);

  // face methods
  int numFaces() const;
  Face3D* addFace(Face3D* face);
  Face3D* addFace(int v0, int v1, int v2);
  Face3D* addFace(int vhandles[3]);
  // update edges
  void updateFaces();

  //gpu operation
  void loadTexture();
  void unloadTexture();

  // vertex methods
  int numVertices() const;
  Vertex3D* addVertex(Vertex3D* vertex);
  Vertex3D* addVertex(const Point3Df& p);
  // update all vertices
  void updateVertices();

  // flags
  void setFlag(int flag, bool value);
  bool getFlag(int flag) const;
  bool hasVertexNormals() const;
  bool hasFaceNormals() const;
  bool hasVertexColors() const;
  bool hasFaceColors() const;
  bool hasVertexFlags() const;
  bool hasFaceFlags() const;
  bool hasTexture() const;
  bool hasImagePose() const;
  bool hasCameraInfo() const;
  bool hasFrameID() const;

  // stream IO
  virtual bool write(OutputHandler &oh) const;
  virtual bool read(InputHandler &ih);
  void write(FILE *fp) const;
  void read(FILE *fp);

  Pose3Df pose;
  int flags;
  VecVertex3D vertices;
  VecFace3D faces;

  //texture mapping
#ifndef NO_GRAPHICS
  Texture texture;
#endif
  Image<Vec3uc> teximage;
  Array2i texmask;
  Pose3Df image_pose;
  sensor_msgs::CameraInfo camera_info;
  std::string frame_id;
};


//==============================================================================
} // NAMESPACE rtc
//==============================================================================
#endif // MESH3D_H_ defined
//==============================================================================
