/*
 * Copyright (C) 2009
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: Autonomous Technologies
 * file .......: rtcOutputHandler.h
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 12/12/2008
 * modified ...: $Date: 2009-03-20 11:36:51 -0700 (Fri, 20 Mar 2009) $
 * changed by .: $Author: wg75pal $
 * revision ...: $Revision: 105 $
 */
#ifndef RTC_OUTPUT_HANDLER_H
#define RTC_OUTPUT_HANDLER_H

//== INCLUDES ==================================================================
#include <string>
#include <iostream>
#include <sstream>

//== NAMESPACES ================================================================
namespace rtc {

/**
 * Base output handler class.
 *
 * The output handler classes offer an abstract interface for the IOObject
 * classes to write() them in different formats.
 *
 * All OutputHandlers are used as wrappers for std::ostream objects.  Therefore,
 * the abstract interface provides the virtual methods use() to indicate which
 * stream has to be used.
 */
class OutputHandler {
public:
  /// default constructor
  OutputHandler();
  /// construct from ostream
  OutputHandler(std::ostream& os);
  /// destructor
  virtual ~OutputHandler();
  /// returns if the data format is binary or ascii
  virtual bool binary() const = 0;
  /// indicate the output stream to be used
  virtual void use(std::ostream& os);
  /// write the name of a name-data pair
  virtual bool writeName(const std::string& name) = 0;
  /// write the separator of a name-data pair
  virtual bool writeSeparator() = 0;
  /// write the EOL symbol
  virtual bool writeEOL() = 0;
  //@{
  /**
   * write members for data
   */
  virtual bool write(const std::string& data) = 0;
  virtual bool write(const double& data) = 0;
  virtual bool write(const float& data) = 0;
  virtual bool write(const int& data) = 0;
  virtual bool write(const unsigned int& data) = 0;
  virtual bool write(const char& data) = 0;
  virtual bool write(const unsigned char& data) = 0;
  virtual bool write(const bool& data) = 0;
  virtual bool write(const long& data) = 0;
  virtual bool write(const unsigned long& data) = 0;
  virtual bool write(const short& data) = 0;
  virtual bool write(const unsigned short& data) = 0;
  virtual bool write(const char* data, long n) = 0;
  //@}
  /// Returns a reference to the stream
  std::ostream& stream();
  /// Check if the state of the stream is good for i/o operations.
  bool good() const;
protected:
  /// pointer to the input stream
  std::ostream* outStream;
};

/**
* handler functions with standard storable interface
*/
template <class T>
bool rtc_write(OutputHandler& oh, const std::string& name, const T& data)
{
  bool res = true;
  res = res && oh.writeName(name);
  res = res && oh.writeSeparator();
  if(oh.binary())
    res = res && oh.write(data);
  else
    oh.stream() << data;
  res = res && oh.writeEOL();
  return res;
};

//==============================================================================
} // namespace rtc
//==============================================================================
#endif // RTC_OUTPUT_HANDLER_H defined
//==============================================================================

