#ifndef RENDERER_H
#define RENDERER_H

//== INCLUDES ==================================================================
#include <list>
#include <string>

#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#define ILUT_USE_OPENGL
#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

#include "rtc/rtcRenderNode.h"

//== NAMESPACES ================================================================
namespace rtc {

//== CONSTANTS ================================================================

typedef enum { GUI_MODE_3D, GUI_MODE_2D } eGUIMode;
typedef enum { IDLE, ROTATING, MOVING, ZOOMING } camera_state_t;

typedef void (*idle_func)(void);
typedef void (*timer_func)(void);
typedef void (*display_func)(void);
typedef void (*keyboard_func)(unsigned char, int, int);
typedef void (*mouse_func)(int, int, int, int);
typedef void (*motion_func)(int, int);
typedef void (*passive_motion_func)(int, int);

typedef struct {
  camera_state_t state;
  float pan, tilt, distance;
  float x_offset, y_offset, z_offset;
  float x_offset_2D, y_offset_2D, rotation_2D, zoom, warp_x, warp_y;
} camera_pose_t, *camera_pose_p;

class RenderManager;

class Renderer {
   friend class RenderManager;
public:
  // c'tor
  Renderer(RenderManager* render_manager,
          int window_x, int window_y,
          int window_width, int window_height,
          const char * title);
  // c'tor
  Renderer();
  // d'tor
  virtual ~Renderer();
  // initialize rendering
  virtual void initialize(RenderManager* render_manager,
                          int window_x, int window_y,
                          int window_width, int window_height,
                          const char * title);
  //clear rendering
  virtual void clear();
  // sets the render mode
  void setRenderMode(eRenderMode render_mode);
  // returns the render mode
  eRenderMode getRenderMode();
  /*****************************************************************************
   * set user defined callback functions
   ****************************************************************************/
  // sets a user defined idle function
  void setIdleFunc(idle_func idle);
  // sets a user defined timer function
  void setTimerFunc(timer_func timer);
  // sets a user defined display function
  void setDisplayFunc(display_func display);
  // sets a user defined keyboard function
  void setKeyboardFunc(keyboard_func keyboard);
  // sets a user defined mouse function
  void setMouseFunc(mouse_func mouse);
  // sets a user defined motion function
  void setMotionFunc(motion_func motion);
  // sets a user defined passive motion function
  void setPassiveMotionFunc(passive_motion_func motion);
  // adds a render node to the renderer
  void addRenderNode(RenderNode* node);
  void clearRenderNode();
  // set 3D camera parameters
  void setCameraParams(double zoom_sensitivity,
                       double rotate_sensitivity,
                       double move_sensitivity,
                       double min_zoom_range,
                       double camera_fov,
                       double min_clip_range,
                       double max_clip_range);
  // set 2D camera parameters
  void set2DCameraParams(double zoom_sensitivity,
                         double rotate_sensivitity,
                         double move_sensitivity);
  // initialize the renderer
  void initialize(int argc, char **argv,
                  int window_x, int window_y,
                  int window_width, int window_height,
                  double fps);
  // set initial camera position
  void setInitialCameraPos(double pan,
                           double tilt,
                           double range,
                           double x_offset,
                           double y_offset,
                           double z_offset);
  // set initial 2D camera position
  void setInitial2DCameraPos(double x_offset,
                             double y_offset,
                             double rotation,
                             double zoom);
  // switch to 2D mode
  void setMode2D(void);
  // switch to 3D mode
  void setMode3D(void);
  // set custom mode
  void setMode(int mode);
  // get current mode
  int getMode(void);
  // force a redraw
  void forceRedraw(void);
  // set the 2D warping
  void set2DWarping(double warp_x, double warp_y);
  // recenter viewer
  void recenter(void);
  // recenter viewer in 2D mode
  void recenter2D(void);
  // pick a point in viewing pane
  void pickPoint(int mouse_x, int mouse_y, double *scene_x, double *scene_y);
  // get 2D position for 2D point
  void get2DPosition(int x, int y, double *xout, double *yout);
  // get 3D position for 3D point
  void get3DPosition(double x, double y, double z, double *xout, double *yout, double *zout);
  // return glut window id
  int getWindowId();
  // returns the camera pose
  camera_pose_p getCameraPose();
  // sets the camera pose
  void setCameraPose(camera_pose_t camera_pose);
  // returns the window width
  int getWindowWidth();
  // returns the window height
  int getWindowHeight();
  // set display to 2d mode
  void setDisplayMode2D(int w, int h);
  // set display to 3d mode
  void setDisplayMode3D(int w, int h, float fov, float zNear, float zFar);
  // switch to 2D mode
  void switchTo2DMode(void);
  // switch to 3D mode
  void switchTo3DMode(void);
  // render bitmap string
  void renderBitmapString(float x, float y, void *font, char *string) ;
  // render 3D bitmap string
  void renderBitmapString3D(float x, float y, float z, void *font, char *string);
  // render bitmap string
  int bitmapStringWidth(void *font, char *string);
  // render centered bitmap string
  void renderBitmapStringCentered(float x, float y, void *font, char *string);
  // take a screenshot
  void takeScreenShot(const char *filename);
  void takeScreenShot();

  void drawCoordinateFrame(double scale = 1.0);
  void drawGrid(double center_x=0, double center_y=0, int size=100, double resolution=1.0) const;
  void drawDashedLine3D(double x1, double y1, double z1, double x2, double y2, double z2, double stripe_len) const;

  Vec3d unProject(double winx, double winy, double winz=0.0);
  void setRasterPos(double winx, double winy, double winz=0.0);

private:
  bool arrow_initialized;
  GLuint arrow_triangle_vertices_vbo;
  GLuint arrow_triangle_normals_vbo;
  GLuint arrow_quad_vertices_vbo;
  GLuint arrow_quad_normals_vbo;
  void initializeArrow();

protected:
  // internal rendering callbacks
  void display(void);
  void mouse(int button, int state, int x, int y);
  void keyboard(unsigned char key, int x, int y);
  void motion(int x, int y);
  void passive_motion(int x, int y);
  void special(int key, int x, int y);
  void visibility(int visible);
  void reshape(int w, int h);
  void timer();
  void idle();

  // internal helper functions
  void rotate_camera(double dx, double dy);
  void zoom_camera(double dy);
  void move_camera(double dx, double dy);
  void rotate_camera_2D(double dx);
  void zoom_camera_2D(double dx);
  void move_camera_2D(double dx, double dy);


public:
  // initialized flag
  bool is_initialized;
  // the rendering mode
  eRenderMode render_mode;
  // list of rendering nodes
  typedef std::vector<RenderNode*> TRenderNodeList;
  TRenderNodeList render_nodes;

  int m_frame;
  int m_time;
  int m_timebase;
  float m_fps;

  int window_id;
  int window_width, window_height;
  float GLUI_x_offset, GLUI_y_offset;
  double fps;
  camera_pose_t camera_pose;
  int last_mouse_x, last_mouse_y;
  bool received_input;

  idle_func user_idle_func;
  timer_func user_timer_func;
  display_func user_display_func;
  keyboard_func user_keyboard_func;
  mouse_func user_mouse_func;
  motion_func user_motion_func;
  passive_motion_func user_passive_motion_func;

  double zoom_sensitivity;
  double rotate_sensitivity;
  double move_sensitivity;
  double min_zoom_range;
  double camera_fov;
  double min_clip_range;
  double max_clip_range;

  double zoom_sensitivity_2D;
  double rotate_sensitivity_2D;
  double move_sensitivity_2D;
  int mode;

  ILuint image_id;
  int screen_shot_number;
};

//==============================================================================
} // namespace rtc
//==============================================================================
#endif // RENDERER_H defined
//==============================================================================

