/*
 * Copyright (C) 2009
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: Autonomous Technologies
 * file .......: rtcBase.cpp
 * authors ....: Benjamin Pitzer, Soeren Kammel
 * organization: Robert Bosch LLC
 * creation ...: 01/21/2009
 * modified ...: $Date: 2009-01-21 18:19:16 -0800 (Wed, 21 Jan 2009) $
 * changed by .: $Author: benjaminpitzer $
 * revision ...: $Revision: 14 $
 */
#include "rtc/rtcBase.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef _MSC_VER
#else
#include <termios.h>
#include <unistd.h>
#endif

//== NAMESPACES ================================================================
namespace rtc {

static FILE *rtc_carp_output = NULL;
static int rtc_carp_verbose = 1;

void
rtc_perror(const char* fmt, ...) {
  va_list args;

  if (rtc_carp_output == NULL)
    rtc_carp_output = stderr;

  va_start(args, fmt);
  vfprintf(rtc_carp_output, fmt, args);
  va_end(args);
  fprintf(rtc_carp_output, "\n");
  fprintf(rtc_carp_output, " : %s\n", strerror(errno));
  fflush(rtc_carp_output);
}

void
rtc_die(const char* fmt, ...) {
  va_list args;

  if (rtc_carp_output == NULL)
    rtc_carp_output = stderr;

  va_start(args, fmt);
  vfprintf(rtc_carp_output, fmt, args);
  va_end(args);
  fprintf(rtc_carp_output, "\n");
  fflush(rtc_carp_output);
  exit(-1);
}

void
rtc_die_syserror(const char* fmt, ...) {
  va_list args;

  if (rtc_carp_output == NULL)
    rtc_carp_output = stderr;

  va_start(args, fmt);
  vfprintf(rtc_carp_output, fmt, args);
  va_end(args);
  fprintf(rtc_carp_output, "\n");
  fprintf(rtc_carp_output, " : %s\n", strerror(errno));
  fflush(rtc_carp_output);

  exit(-1);
}

void
rtc_warn(const char* fmt, ...) {
  va_list args;

  if (rtc_carp_output == NULL)
    rtc_carp_output = stderr;

  va_start(args, fmt);
  vfprintf(rtc_carp_output, fmt, args);
  va_end(args);
  fprintf(rtc_carp_output, "\n");
  fflush(rtc_carp_output);
}

void
rtc_verbose(const char *fmt, ...) {
  va_list args;

  if (!rtc_carp_verbose)
    return;

  if (rtc_carp_output == NULL)
    rtc_carp_output = stderr;

  va_start(args, fmt);
  vfprintf(rtc_carp_output, fmt, args);
  va_end(args);
  fprintf(rtc_carp_output, "\n");
  fflush(rtc_carp_output);
}

void
rtc_carp_set_output(FILE *output) {
  rtc_carp_output = output;
}

void
rtc_carp_set_verbose(int verbosity) {
  rtc_carp_verbose = verbosity;
}

int
rtc_carp_get_verbose(void) {
  return rtc_carp_verbose;
}

bool rtc_find_commandline_parameter(int argc, char **argv, const char* param, char* param_value) {
  int i;

  if (argc < 2)
    return false;

  for (i = 1; i < argc-1; i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(argv[i]+1, param)) {
        strcpy(param_value, argv[i+1]);
        return true;
      }
    }
  }

  return false;
}

bool rtc_find_commandline_parameter(int argc, char **argv, const char* param) {
  int i;

  if (argc < 2)
    return false;

  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(argv[i]+1, param)) {
        return true;
      }
    }
  }

  return false;
}

/* This function displays a progress bar
   It uses a 1:2 scales. */
void rtc_progress_bar_print(double current, double total)
{
  int i;
  int per_tot, per_cur, per_tot_scl, per_cur_scl;

  per_tot = 100;
  /* total percentage */
  if (total==0.0) total=1;
  per_cur = (per_tot * current) / total;
  /* current percentage */

  /* 1:2 scale */
  per_tot_scl = per_tot / 2;
  per_cur_scl = per_cur / 2;

  /* Refresh ther bar indicator */
  printf("\r [");

  /* Display the bar indicator:
     - =   : indicates a part the bar
     - >   : indicates the end of the bar
             (if it's finish, convert '>' to '='*/
  for (i = 0; i < per_cur_scl; i++)
    if ((i + 1) == per_cur_scl)
      if (per_cur_scl == 50)
        putchar('=');
      else
        putchar('>');
    else
      putchar('=');

  for (i = per_cur_scl; i < per_tot_scl; i++)
    putchar(' ');

  printf("] %3d%% | ", per_cur);
  fflush(stdout);
}

//==============================================================================
} // namespace rtc
//==============================================================================
