/*
* Copyright (C) 2007
* Robert Bosch Corporation
* Research and Technology Center North America
* Palo Alto, California
*
* All rights reserved.
*/

/*--------------------------------------------------------------------
* project ....: PUMA: Probablistic Unsupervised Model Acquisition
* file .......: RenderNode.cpp
* authors ....: Benjamin Pitzer
* organization: Bosch, RTC
* creation ...: 16.04.2007
* revisions ..: $Id: RenderNode.cpp,v 1.2 2007/05/08 18:54:45 prb2pal Exp $
*/

//== INCLUDES =================================================================
#include "rtc/rtcRenderer.h"
#include <GL/gl.h>
#include "rtc/rtcRenderNode.h"

//== NAMESPACES ===============================================================

namespace rtc {

//== IMPLEMENTATION ==========================================================

// --------------------------------------------------
// RenderNode::Parameters
// --------------------------------------------------

// default constructor
RenderNode::Parameters::Parameters()
{
  draw_faces = true;
  draw_vertices = false;
  draw_wireframe = false;
  use_vertex_colors = false;
  color_mode = NO_COLOR;
  highlite = false;
}

// copy constructor
RenderNode::Parameters::Parameters(const Parameters& other)
{
  copy(other);
}

// destructor
RenderNode::Parameters::~Parameters()
{
}

// copy member
RenderNode::Parameters&
  RenderNode::Parameters::copy(const Parameters& other)
{
  draw_faces = other.draw_faces;
  draw_vertices = other.draw_vertices;
  draw_wireframe = other.draw_wireframe;
  use_vertex_colors = other.use_vertex_colors;
  color_mode = other.color_mode;
  highlite = other.highlite;
  return *this;
}

// alias for copy member
RenderNode::Parameters&
  RenderNode::Parameters::operator=(const Parameters& other)
{
  return copy(other);
}

// clone member
RenderNode::Parameters*
  RenderNode::Parameters::clone() const
{
  return new Parameters(*this);
}

/*
* read the RenderNode parameters
*/
bool RenderNode::Parameters::read(std::istream& is)
{
  return true;
}

/*
* write the RenderNode parameters
*/
bool RenderNode::Parameters::write(std::ostream& os) const
{
  return true;
}

// --------------------------------------------------
// RenderNode
// --------------------------------------------------

// constructor
RenderNode::RenderNode(Renderer* renderer)
{
  m_params = new Parameters();
  m_is_initialized = false;
  m_renderer = renderer;
}

// destructor
RenderNode::~RenderNode()
{
  if(m_params) delete m_params;
}

/* sets the Renderer's Parameters.
The Renderer keeps its own copy of the given Parameters.*/
bool RenderNode::setParameters(const RenderNode::Parameters& theParams)
{
  if(m_params) delete m_params;
  m_params = 0;
  m_params = theParams.clone();
  return true;
}

/*
* returns current Parameters.
*/
const RenderNode::Parameters& RenderNode::getParameters() const
{
  return *m_params;
}

/*
* returns current Parameters.
*/
RenderNode::Parameters& RenderNode::getParameters()
{
  return *m_params;
}

void RenderNode::initialize()
{
}

void RenderNode::rebuild()
{
  clear();
  initialize();
}

void RenderNode::clear()
{
}

void RenderNode::render()
{
}

void RenderNode::preRender()
{
  // apply transform
  glPushMatrix();
  Transformf t(m_transform.transposed());
  glMultMatrixf(&t.x[0]);

  if(m_params->highlite)
    glColor3f(1.0,1.0,0.0); // yellow
  else {
    switch(m_params->color_mode)
    {
    case Parameters::NO_COLOR:
      break;
    case Parameters::VERTEX_COLOR:
      //glColor3f(0.705f, 0.976f, 0.270f); // greenish
      glColor3f(1.0,1.0,1.0); // white
      break;
    case Parameters::TEXTURE_COLOR:
      glColor3f(1,1,1);
      break;
    default:
      glColor3f(1,1,1);
      break;
    }
  }

}

void RenderNode::postRender()
{
  // reset transformation matrix
  glPopMatrix();
}


void RenderNode::setTransform(Transformf transform)
{
  m_transform = transform;
}

const Transformf& RenderNode::getTransform()
{
  return m_transform;
}

eRenderMode RenderNode::getRenderMode()
{
  return m_renderer->getRenderMode();
}

void RenderNode::displayFaces(bool b)
{
  m_params->draw_faces = b;
}

void RenderNode::displayVertices(bool b)
{
  m_params->draw_vertices = b;
}

void RenderNode::displayWireframe(bool b)
{
  m_params->draw_wireframe = b;
}

void RenderNode::setColorMode(Parameters::ColorMode b)
{
  m_params->color_mode = b;
}

void RenderNode::setHighlite(bool b)
{
  m_params->highlite = b;
}


//=============================================================================
} // namespace puma
//=============================================================================
