/*
 * Copyright (C) 2007
 * Robert Bosch LLC
 * Research and Technology Center North America
 * Palo Alto, California
 *
 * All rights reserved.
 *
 *------------------------------------------------------------------------------
 * project ....: PUMA: Probablistic Unsupervised Model Acquisition
 * file .......: mesh.cpp
 * authors ....: Benjamin Pitzer
 * organization: Robert Bosch LLC
 * creation ...: 03/28/2008
 * modified ...: $Date: 2008-12-19 10:32:32 -0800 (Fri, 19 Dec 2008) $
 * changed by .: $Author: wg75pal $
 * revision ...: $Revision: 615 $
 */

//== INCLUDES ==================================================================
#include "rtc/rtcTriMesh.h"

//== NAMESPACES ================================================================
namespace rtc {

Vec3f point_to_vec(const Point& p)
{
  return(Vec3f(p[0],p[1],p[2]));
}

Vec3f normal_to_vec(const Normal& p)
{
  return(Vec3f(p[0],p[1],p[2]));
}

Vec3uc color_to_vec(const Color& p)
{
  return(Vec3uc(p[0],p[1],p[2]));
}

Point vec_to_point(const Vec3f& v)
{
  return(Point(v[0],v[1],v[2]));
}

Normal vec_to_normal(const Vec3f& v)
{
  return(Normal(v[0],v[1],v[2]));
}

Color vec_to_color(const Vec3uc& v)
{
  return(Color(v[0],v[1],v[2]));
}

void convertMeshToTriMesh(const Mesh3D& input, TriMesh& output)
{
  output.request_vertex_normals();
  output.request_face_normals();
  output.request_vertex_colors();

  //== Vertices ==========================================================
  int nverts = (int)input.vertices.size();
  for (int i=0;i<nverts;++i)
  {
    Vec3f& p = input.vertices[i]->p;
    Vec3f& n = input.vertices[i]->n;
    Vec3uc& c = input.vertices[i]->c;
    VertexHandle vh = output.add_vertex(vec_to_point(p));
    output.set_normal(vh,vec_to_normal(n));
    output.set_color(vh,vec_to_color(c));
  }

  //== Faces ==========================================================
  int nfaces = (int)input.faces.size();
  std::vector<VertexHandle> newFace(3);
  for (int i=0;i<nfaces;++i)
  {
    Face3D* f = input.faces[i];
    newFace[0] = output.vertex_handle(f->v(0));
    newFace[1] = output.vertex_handle(f->v(1));
    newFace[2] = output.vertex_handle(f->v(2));
    FaceHandle fh = output.add_face(newFace);
    output.set_normal(fh,vec_to_normal(f->n));
  }
}

void convertTriMeshToMesh(const TriMesh& input, Mesh3D& output)
{
  //== Vertices ==========================================================
  ConstVertexIter vIt, vEnd;
  ConstFaceIter fIt, fEnd;
  ConstFaceVertexIter cfvIt;

  // clean first
  output.clear();

  int i = 0;
  for(vIt=input.vertices_begin(), vEnd = input.vertices_end(); vIt != vEnd; ++vIt,++i) {
    Vec3f p = point_to_vec(input.point(vIt));
    Vertex3D* v = output.addVertex(p);
    v->n = normal_to_vec(input.normal(vIt));
    v->c = color_to_vec(input.color(vIt));
    v->setBoundary(input.is_boundary(vIt));
  }

  int face[3];
  for(fIt=input.faces_begin(), fEnd = input.faces_end(); fIt != fEnd; ++fIt) {
    // get triangle vertices
    cfvIt = input.cfv_iter(fIt);
    for(int j=0; j<3; j++) {
      face[j] = cfvIt.handle().idx();
      ++cfvIt;
    }
    Face3D* f = output.addFace(face);
    f->n = normal_to_vec(input.normal(cfvIt));
  }
}


//==============================================================================
} // NAMESPACE puma
//==============================================================================
